#===============================================================================
# CERN (BE-CO-HT)
# MultiBoot example script
#===============================================================================
# author: Theodor Stana (t.stana@cern.ch)
#
# date of creation: 2013-08-13
#
# version: 1.0
#
# description:
#     This module exemplifies reading and writing to the CONV-TTL-BLO FPGA
#     registers for sending a bitstream to the CONV-TTL-BLO flash and then
#     issuing the IPROG command to the FPGA to load the new bitstream from the
#     flash.
#
#     To communicate to the CONV-TTL-BLO, an EI2C object is used to open a
#     coonection to the ELMA crate and send the writereg, writemregs and readreg
#     commands.
#
# dependencies:
#     ei2c.py
#     ei2cexcept.py
#
# references:
#   [1] M25P32 32Mb 3V NOR Serial Flash Embedded Memory
#       http://cern.ch/go/z7pw
#
#===============================================================================
# GNU LESSER GENERAL PUBLIC LICENSE
#===============================================================================
# This source file is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation; either version 2.1 of the License, or (at your
# option) any later version. This source is distributed in the hope that it
# will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Lesser General Public License for more details. You should have
# received a copy of the GNU Lesser General Public License along with this
# source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
#===============================================================================
# last changes:
#    2013-08-13    Theodor Stana     File created
#===============================================================================
#  TODO: -
#===============================================================================

import random
import sys
import time
sys.path.append("../ei2c")
from ei2c import *
from ei2cexcept import *
from functools import partial

#===============================================================================
MB_BASE = 0x40
MB_CR_OFS = 0x00
MB_SR_OFS = 0x04
MB_GBBAR_OFS = 0x08
MB_MBBAR_OFS = 0x0C
MB_FAR_OFS   = 0x10
#===============================================================================

#===============================================================================
slot = 0
#===============================================================================

# This function handles the OR-ing together of the data bytes and the control
# byte in FAR.
#
# Formatting of data bytes in the FAR register should be done outside this
# function.
def spi_transfer(nbytes, cs, dat):
  retval = 0
  wval = []

  # control byte, to be shifted left by 24
  ctrl = ((cs << 3) | 0x4) | (nbytes-1)

  # Use appropriate command by type
  #
  # write - we send up to three data bytes in one FAR register write
  # writemregs - we send up to 24 data bytes in eight FAR register writes
  if isinstance(dat,int):
    testelma.write(slot, MB_BASE+MB_FAR_OFS, (ctrl << 24) | dat)
  else:
    for i in xrange(len(dat)):
      wval.append((ctrl << 24) | dat[i])
    testelma.writemregs(slot, MB_BASE+MB_FAR_OFS, wval)

  # Read the data and prepare the return value
  while (retval & (1 << 28) == 0):
    retval = testelma.read(slot, MB_BASE+MB_FAR_OFS)

  return retval & 0xFFFFFF

# This function is used to write data bytes to flash. It calls spi_transfer
# and sends the commands to the flash chip. You can follow the logic of this
# function by looking at the sequence of the write command in [1]
def flash_write(addr, dat):

  # write enable cmd
  spi_transfer(1,1,0x06)
  spi_transfer(1,0,0)

  # write page cmd
  spi_transfer(1,1,0x02)

  # send address in reverse order
  addr = ((addr & 0xff0000) >> 16) | (((addr & 0xff00) >> 8) << 8) | ((addr & 0xff) << 16)
  spi_transfer(3,1,addr)

  # Now, massage the data array into 3-byte commands sent into 32-bit register,
  # in batches of eight registers (writemregs command).
  i = 0
  l = len(dat)
  while (l):
    wval = []
    # If we have more than 24 bytes left in the data array, we can use the full
    # writemregs
    if (int(l/24)):
      for j in xrange(0, 8):
        wval.append((dat[(i+2)+3*j] << 16) | (dat[(i+1)+3*j] << 8) | dat[i+3*j])
      spi_transfer(3, 1, wval)
      i += 24;
      l -= 24;

    # When we're left to fewer than 24 bytes in the array, try to use one more
    # writemregs with as many regs as possible
    elif (int(l/3)):
      for j in xrange(0, l/3):
        wval.append((dat[(i+2)+3*j] << 16) | (dat[(i+1)+3*j] << 8) | dat[i+3*j])
      spi_transfer(3,1,wval);
      i += 3 * int(l/3)
      l -= 3 * int(l/3)

    # When we're down to less than three bytes, just use simple one-byte xfers
    elif (l):
      for j in xrange(0, l):
        spi_transfer(1,1,dat[i])
        i += 1
        l -= 1

  spi_transfer(1,0,0)


# This function is used to read a number of data bytes from the flash memory
# It returns the values of the three consecutive flash registers packed into
# one 24-bit data value in little-endian order
def flash_read(addr, nrbytes):
  ret = []
  spi_transfer(1,1,0x0b)

  # send address in reverse order
  addr = ((addr & 0xff0000) >> 16) | (((addr & 0xff00) >> 8) << 8) | ((addr & 0xff) << 16)
  spi_transfer(3,1,addr)
  spi_transfer(1,1,0)

  # Read bytes in groups of three
  for i in range(nrbytes):
    ret.append(spi_transfer(1,1,0))
  spi_transfer(1,0,0)

  return ret

# Send a sector erase command
def flash_serase(addr):
  spi_transfer(1,1,0x06)
  spi_transfer(1,0,0)
  spi_transfer(1,1,0xD8)
  # send address in reverse order
  addr = ((addr & 0xff0000) >> 16) | (((addr & 0xff00) >> 8) << 8) | ((addr & 0xff) << 16)
  spi_transfer(3,1,addr)
  spi_transfer(1,0,0)

# Send a block erase command
def flash_berase():
  spi_transfer(1,1,0x06)
  spi_transfer(1,0,0)
  spi_transfer(1,1,0xc7)
  spi_transfer(1,0,0)

# Read status register command
def flash_rsr():
  spi_transfer(1,1,0x05)
  ret = spi_transfer(1,1,0)
  spi_transfer(1,0,0)
  return ret


#=============================================================================#

#
# Read from flash
#
def _read(addr):
  print "Reading 256 bytes from 0x%06X" % addr
  tr0 = time.time()
  rd = flash_read(addr,256)
  tr1 = time.time()
  rd = "".join(["%02X" % b for b in rd])
  print rd

#
# Write to flash
#
def _write(addr):

  # Ookay, now we start to write the flash
  print "Writing bitstream..."

  # Prepare time arrays, to calculate average time each operation takes
  tdat = []
  twr  = []
  twa  = []
  te = []
  tp = []

  # Ask for and open bitstream file
  fname = raw_input("Input file name: ")
  f = open(fname,'rb')

  tw0 = time.time()

  # Each line in the input file contains the data for one page
  for dat in iter(partial(f.read, 256), ''):
    dat = [int(d.encode("hex"), 16) for d in dat]
    print "%x" % addr

    # Erase on sector boundary
    if not (addr % 0x10000):
      print 'erase'
      t1 = time.time()
      flash_serase(addr)
      while (flash_rsr() & 0x01):
        pass
      t2 = time.time()
      te.append(t2-t1)

    t3 = time.time()
    flash_write(addr, dat)
    t4 = time.time()
    while (flash_rsr() & 0x01):
      pass
    t5 = time.time()
    addr += 256
    twr.append(t4-t3)
    twa.append(t5-t4)
    tp.append(t5-t3)

  tw1 = time.time()

  # Close file handle
  f.close()

  # Phew! We're ready, so calculate the mean time it took for each process
  print "DONE!"
  print "erase time: %2.6f" % float(sum(te)/len(te))
  print "write time: %2.6f" % float(sum(twr)/len(twr)) #(t4-t3)
  print "wait time:  %2.6f" % float(sum(twa)/len(twa)) #(t5-t4)
  print "page write time: %2.6f" % float(sum(tp)/len(tp))
  print "total write time: %2.6f" % float(tw1-tw0)

#
# Start IPROG sequence
#
def _iprog(addr):
  print "Issuing IPROG command..."
  testelma.write(slot, MB_BASE+MB_GBBAR_OFS, 0x44 | (0x0b << 24))
  testelma.write(slot, MB_BASE+MB_MBBAR_OFS, addr | (0x0b << 24))
  testelma.write(slot, MB_BASE+MB_CR_OFS, 0x10000)
  try:
    testelma.write(slot, MB_BASE+MB_CR_OFS, 0x20000)
  except NAckError:
    pass

  # Set timeout...
  t0 = time.time()
  t1 = t0 + 20

  # and wait for the FPGA to gracefully respond, or die trying
  while (1):
    try:
      if (time.time() >= t1):
        print "Timeout, IPROG unsuccessful!"
        break
      if ((testelma.read(slot, 0x4) & 0xFF) == 0x00):
        print "IPROG unsuccessful, fallback to Golden bitstream occured!"
        break
      else:
        print "IPROG successful!"
        break
    except NAckError:
        continue

def _rdcfgreg():
  print "Press 'q' to end config reg readout"
  while 1:
    try:
      reg = raw_input('Address (hex): ')
      reg = int(reg, 16)
      if (reg < 0x00) or (reg > 0x22):
        raise ValueError
      testelma.write(slot, MB_BASE+MB_CR_OFS, (1 << 6) | reg)
      val = testelma.read(slot, MB_BASE+MB_SR_OFS)
      if (val & (1 << 16)):
        print "REG(0x%02X) = 0x%04X" % (reg, val & 0xFFFF)
      else:
        print "Status invalid!"
    except ValueError:
      if (reg == 'q'):
        break
      print "Please input a hex value in the range [0x00, 0x22] or 'q' to quit"



#=============================================================================#



# MAIN "FUNCTION"
if __name__ == "__main__":
  # Ask for username and password and retry init in case of error
  while 1:
    ip   = raw_input("Crate IP or hostname: ")
    user = raw_input("Username: ")
    pwd  = raw_input("Password: ")
    try:
      testelma = EI2C(ip, user, pwd)
      testelma.open()
      break
    except BadHostnameError as e:
      print e.strerror
    except BadUsernameError as e:
      print e.strerror
    except BadPasswordError as e:
      print e.strerror


  # Wait for proper slot number
  while 1:
    try:
      slot = raw_input("Slot no.: ")
      slot = int(slot)
      break
    except TypeError as e:
      print "Please input a decimal slot number."
    except SlotError as e:
      print e.strerror
    except KeyboardInterrupt:
      sys.exit();
    except:
      print "Unexpected error: ", sys.exc_info()[0]


  # The PULSETEST bitstream has the MultiBoot module starting at address 0x300
  if (testelma.read(slot, 0x04) & 0xFF == 0x99):
    MB_BASE = 0x300;

  # Read config reg, if user wants to
  while 1:
    rdcfgreg = raw_input("Read config reg? (y/n) ")
    if (rdcfgreg == 'y'):
      _rdcfgreg()
      break
    elif (rdcfgreg == 'n'):
      break

  # Ask else to do (read/write/iprog)
  while 1:
    rd = raw_input("Read first page? (y/n) ")
    if (rd == 'y') or (rd == 'n'):
      break

  while 1:
    wr = raw_input("Write to flash?  (y/n) ")
    if (wr == 'y'):
      while 1:
        wr = raw_input("Are you sure?    (y/n) ")
        if (wr == 'y') or (wr == 'n'):
          break
      break
    elif (wr == 'n'):
      break

  while 1:
    iprog = raw_input("Issue IPROG?     (y/n) ")
    if (iprog == 'y') or (iprog == 'n'):
      break

  # Ask for multiboot address
  while 1:
    try:
      multiboot_addr = raw_input("MultiBoot bitstream address (hex): ")
      multiboot_addr = int(multiboot_addr, 16)
      print '0x%06x' % multiboot_addr
      break
    except ValueError as e:
      print "Please input a hexadecimal address!"
    except KeyboardInterrupt:
      sys.exit()
    except:
      print "Unexpected error: ", sys.exc_info()[0]


  # Now do something based on user selection

  # read from flash
  if (rd == 'y'):
    _read(multiboot_addr)

  # write to flash
  if (wr == 'y'):
    _write(multiboot_addr)

  # Finally, issue the reprogramming (IPROG) command
  if (iprog == 'y'):
    _iprog(multiboot_addr)

  # Close I2C connection
  testelma.close()

